// created on 6/14/2004 at 10:39 PM
/*
 *   Copyright (c) 2004, Alexandros Frantzis (alf82 [at] freemail [dot] gr)
 *
 *   This file is part of Bless.
 *
 *   Bless is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   Bless is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Bless; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
using Bless.Gui.Drawers;
using Bless.Util; 
 
namespace Bless.Gui.Areas {

public abstract class Area {

	protected Gtk.DrawingArea drawingArea;
	protected Drawer font;
	protected ByteBuffer byteBuffer;
	
	// display
	protected int x;
	protected int y;
	protected int width;
	protected int height;
	protected long offset;
	protected int bpr;
	protected string type;
	protected int dpb; // digits per byte 
	protected Gdk.Pixmap backPixmap;
	
	// GC's
	protected Gdk.GC backEvenGC;
	protected Gdk.GC backOddGC;
	protected Gdk.GC backrEvencSelGC;
	protected Gdk.GC backrOddcSelGC;
	protected Gdk.GC cursorGC;
	
	protected Range selection;
	
	// cursor
	protected long cursorOffset;
	protected int  cursorDigit;
	protected bool cursorFocus;

	// track changes
	protected Range prevSelection;  
	protected long prevOffset; 
	protected long prevCursorOffset;
	protected int  prevCursorDigit;
	
	// Abstract methods
	//
	abstract protected void RenderRow(int i, int p, int n, bool blank, bool select);
	
	abstract public void GetDisplayInfoByOffset(long off, out int orow, out int obyte, out int ox, out int oy);
	
	public enum GetOffsetFlags { Eof=1, Abyss=2}
	
	abstract public long GetOffsetByDisplayInfo(int x, int y,out int digit, out GetOffsetFlags rflags);
	
	virtual public bool HandleKey(Gdk.Key key, bool overwrite)
	{
		return false;
	}
	

	///<summary>
	/// Calculates the width in pixels the    
	/// area will occupy with n bytes per row.
	///</summary>
	abstract public int CalcWidth(int n);
	
	
	static public Area Factory(string name, Gtk.DrawingArea da)
	{
		Area a;
		
		switch (name) {
			case "hexadecimal":
				a=new HexArea(da);
				break;
			case "binary":
				a=new BinaryArea(da);
				break;
			case "decimal":
				a=new DecimalArea(da);
				break;
			case "octal":
				a=new OctalArea(da);
				break;
			case "ascii":
				a=new AsciiArea(da);
				break;
			case "separator":
				a=new SeparatorArea(da);
				break;
			case "offset":
				a=new OffsetArea(da);
				break;				
			default:
				a=null;
				break;
		}	
		
		return a;
	
	}

	public Area(Gtk.DrawingArea da) 
	{
		drawingArea=da;
		backPixmap=new Gdk.Pixmap(da.GdkWindow, 1,1,-1);
		backEvenGC=new Gdk.GC(da.GdkWindow);
		backOddGC=new Gdk.GC(da.GdkWindow);
		backrOddcSelGC=new Gdk.GC(da.GdkWindow);
		backrEvencSelGC=new Gdk.GC(da.GdkWindow);
		cursorGC=new Gdk.GC(da.GdkWindow);
		
		selection=new Range();
		prevSelection=new Range();
		
		//InvalidateScrolling();
	}
	
	// Properties
	//
	public int X {
		set { x=value; }
	 	get	{ return x; }
	} 

	public int Y {
		set { y=value; }
	 	get	{ return y; }
	}
	
	public int Width {
		set {
			// no need to allocate new memory if no change has happened
			if (value==width)
				return;
			width=value;
			if (width!=0 && height!=0) {
				backPixmap.Dispose();
				backPixmap=new Gdk.Pixmap(drawingArea.GdkWindow,width,height,-1);
			}
		}
	 	get	{ return width; }
	} 

	public int Height {
		set {
			// no need to allocate new memory if no change has happened
			if (value==height)
				return;
			height=value;
			if (width!=0 && height!=0) {
				backPixmap.Dispose(); 
				backPixmap=new Gdk.Pixmap(drawingArea.GdkWindow,width,height,-1);
			} 
		}
	 	get	{ return height; }
	}
	
	public int BytesPerRow {
		set { bpr=value; }
	 	get	{ return bpr; }
	}
		
	public int DigitsPerByte {
	 	get	{ return dpb; }
	}
	
	public long Offset {
		set { prevOffset=offset; offset=value;} 
		get { return offset;}
	}

	public long CursorOffset {
		set { prevCursorOffset=cursorOffset; cursorOffset=value;} 
		get { return cursorOffset;}
	}
	
	public long PrevCursorOffset { 
		get { return prevCursorOffset;}
	}
	
	public int CursorDigit {
		set { prevCursorDigit=cursorDigit; cursorDigit=value;} 
		get { return cursorDigit;}
	}
	
	public bool HasCursorFocus {
		set { cursorFocus=value; CursorDigit=-1;} 
		get { return cursorFocus;}
	}
	
	public Drawer Font {
		get { return font; }
		set { 
			font=value; 
			Gdk.Color col=new Gdk.Color();
			Gdk.Color.Parse(font.Info.rEvencEvenBg, ref col); 
			backEvenGC.RgbFgColor=col;
			Gdk.Color.Parse(font.Info.rOddcEvenBg, ref col); 
			backOddGC.RgbFgColor=col;
			Gdk.Color.Parse(font.Info.rEvencSelBg, ref col); 
			backrEvencSelGC.RgbFgColor=col;
			Gdk.Color.Parse(font.Info.rOddcSelBg, ref col); 
			backrOddcSelGC.RgbFgColor=col;
			Gdk.Color.Parse("red", ref col); 
			cursorGC.RgbFgColor=col;
		}
	}

	public string Type {
		get { return type; }
	}
	
	public Gtk.DrawingArea DrawingArea {
		get { return drawingArea; }
	}
	
	public ByteBuffer Buffer {
		get { return byteBuffer; }
		set { byteBuffer=value; }
	}
	
	public Range Selection {
		get { return selection; }
		set { selection=value; }
	} 
	
	protected void InvalidateScrolling() 
	{
		prevOffset=long.MaxValue;
	}
	
	///<summary>Draws the area</summary>
	virtual public void Render(bool force) 
	{
		Gdk.Window win=drawingArea.GdkWindow;
		
		// if we don't have enough space to draw return 
		if (bpr<=0)
			return;
		
		// 0. Find what has changed
		bool selectionChanged;
		bool offsetChanged;
		bool cursorChanged;
		
		if (!selection.Equals(prevSelection))
			selectionChanged=true;
		else		
			selectionChanged=false;
	
		if (prevOffset != offset) 
			offsetChanged=true;
		else
			offsetChanged=false;
			
		if (prevCursorOffset != cursorOffset || prevCursorDigit != cursorDigit) 
			cursorChanged=true;
		else
			cursorChanged=false;	
			
		// if nothing has changed and we are not forced to
		// redraw just return
		if (!offsetChanged && !selectionChanged && !cursorChanged && !force)
			return;	
	
		// 1. find out number of rows, bytes etc
		//
		int nrows=height/font.Height;
		long bleft=nrows*bpr;
        int rfull=0;
		int blast=0;

		if (bleft+offset >= byteBuffer.Size)
			bleft=byteBuffer.Size - offset;
                                                                                
		// calculate number of full rows
		// and number of bytes in last (non-full)
		rfull=(int)(bleft/bpr);
		blast=(int)(bleft%bpr);
		
		
		// Case 1: Only cursor has changed
		//
		if (cursorChanged && !selectionChanged && !offsetChanged && !force) {
			// Redraw byte at previous pos if it is visible
			if (prevCursorOffset >= offset && prevCursorOffset < offset+bleft) {	 
				int pcRow, pcByte, pcX, pcY;
 				GetDisplayInfoByOffset(prevCursorOffset, out pcRow, out pcByte, out pcX, out pcY);
 				RenderRow(pcRow, pcByte, 1, false, selection.Contains(prevCursorOffset));	
 			}
 			// previous pos was one past the end of buffer
 			else if (prevCursorOffset == byteBuffer.Size && prevCursorOffset == offset+bleft) {
 				int pcRow, pcByte, pcX, pcY;
 				GetDisplayInfoByOffset(prevCursorOffset, out pcRow, out pcByte, out pcX, out pcY);
 				backPixmap.DrawRectangle(backEvenGC, true, pcX, pcY, font.Width*dpb, font.Height);
 			}
 			
 			if ((cursorOffset >= offset && cursorOffset < offset+bleft)
 				|| (cursorOffset == byteBuffer.Size && cursorOffset == offset+bleft))
				RenderCursor();
				
 			prevCursorOffset=cursorOffset;
 			prevCursorDigit=cursorDigit;
 			win.DrawDrawable(backEvenGC, backPixmap, 0, 0, x, y, width, height);
 			return;
		}
				
		// Case 2: Only the selection has changed
		//
		if (selectionChanged && !offsetChanged && !force) {
			Range clip;
			
			// make sure we get an empty clipping Range when bleft==0
			if (bleft>0)
				clip=new Range(offset, offset+bleft-1);
			else
				clip=new Range();
				
			RenderSelectionDiffs(clip);
			// update selection tracking
			prevSelection.Start=selection.Start;
			prevSelection.End=selection.End;
			
			
			// Redraw byte at previous pos if it is visible
			if (prevCursorOffset >= offset && prevCursorOffset < offset+bleft) {	 
				int pcRow, pcByte, pcX, pcY;
 				GetDisplayInfoByOffset(prevCursorOffset, out pcRow, out pcByte, out pcX, out pcY);
 				RenderRow(pcRow, pcByte, 1, false, selection.Contains(prevCursorOffset));	
 			}
			// previous cursor pos was one past the end of buffer
			// clear it (because it is not overwritten by the selection)
 			else if (prevCursorOffset == byteBuffer.Size && prevCursorOffset == offset+bleft) {
 				int pcRow, pcByte, pcX, pcY;
 				GetDisplayInfoByOffset(prevCursorOffset, out pcRow, out pcByte, out pcX, out pcY);
 				backPixmap.DrawRectangle(backEvenGC, true, pcX, pcY, font.Width*dpb, font.Height);
 			}
 			
			if ((cursorOffset >= offset && cursorOffset < offset+bleft)
 				|| (cursorOffset == byteBuffer.Size && cursorOffset == offset+bleft))
				RenderCursor();
			// draw and return
			win.DrawDrawable(backEvenGC, backPixmap, 0, 0, x, y, width, height);
			return;
		}
		
		// update changes tracking
		prevSelection.Start=selection.Start;
		prevSelection.End=selection.End;
		prevCursorOffset=cursorOffset;
		prevCursorDigit=cursorDigit;
		prevOffset=offset;
		
		// Case 3: Offset changed
		//
		
		// draw the area background
		backPixmap.DrawRectangle(backEvenGC, true, 0,0, width, height);
		
                                                                                
		// 2. Find out which rows are fully affected by the selection
		//	
		Range rSel=new Range(selection);
		rSel.Sort();
		Range r1=new Range(offset, offset+bleft-1);
		r1.Intersect(rSel);
		int startSelRow=(int)((r1.Start-offset)/bpr);
		int endSelRow=(int)((r1.End-offset)/bpr);
		startSelRow++; // first and last row are
		endSelRow--; // not considered fully affected
		
		
		int i=0;
		
		// TO-DO: scolling should take selection into account 
		/*if (false) {
			long diff=(offset-prevOffset)/bpr;
			//System.Console.WriteLine("Scroll {0}-{1}={2}",offset, prevOffset, diff);
			int sign=1;
			if (diff<0) {
				diff=-diff;
				sign=-1;
			}
			
			if (diff<nrows) {
				if (sign==1) {
					int dh=(int)diff*font.Height;
					int ry=y+dh;
					backPixmap.DrawDrawable(backEvenGC, win, x, ry, 0, 0, width, (nrows-(int)diff)*font.Height);
					i=nrows-(int)diff;
					
				}
				else {
					int dh=(int)diff*font.Height;
					int ry=y+dh;
					backPixmap.DrawDrawable(backEvenGC, win, x, y, 0, dh, width, (nrows-(int)diff)*font.Height);
					rfull=(int)diff;
				}
			}
			
		}*/
 		
		// 3. Render normal rows and half-selected rows
 		//
 		
 		// draw full rows not affected by selection
  		for (; i<rfull; i++) {
 			if (i < startSelRow || i> endSelRow)
 				RenderRow(i, 0, bpr, true, false);
  		}
                                                                               
 		// draw last if not affected by selection
 		if (blast!=0 && (i < startSelRow || i> endSelRow))
  			RenderRow(i, 0, blast, true, false);
  		
 		// 4. Render selection
 		//
 		if (!r1.IsEmpty())
 			RenderSelection(r1);	
 			
 		if ((cursorOffset >= offset && cursorOffset < offset+bleft)
 			|| (cursorOffset == byteBuffer.Size && cursorOffset == offset+bleft))
			RenderCursor();
			
  		win.DrawDrawable(backEvenGC, backPixmap, 0, 0, x, y, width, height);
  	}
  	
	///<summary>
	/// Render the current selection taking
	/// into account differences from the previous one.
	///</summary>  	
  	protected void RenderSelectionDiffs(Range clip)
  	{
		// find what parts of the selection have changed
		Range added=new Range(selection);
		added.Sort();
		
		// find unchanged range		
		Range common=new Range(prevSelection);
		common.Sort();
		common.Intersect(added);
		common.Intersect(clip);
			
		// find lost (now unselected) ranges
		Range lost1=new Range(prevSelection);
		Range lost2=new Range();
		lost1.Sort();
		lost1.Difference(added, lost2);
		lost1.Intersect(clip);
		lost2.Intersect(clip);
			
		// find added ranges
		Range old=new Range(prevSelection);
		Range added1=new Range();
		old.Sort();
		added.Difference(old, added1);
		added.Intersect(clip);
		
		//System.Console.WriteLine("RenderSelectionDiffs: Clip=({0},{1}) Added=({2},{3}) Lost1=({4},{5}) Lost2=({6},{7}) Common=({8},{9}) PrevSel=({10},{11}) CurSel=({12},{13})",
		//						clip.Start, clip.End, added.Start, added.End, lost1.Start, lost1.End, lost2.Start, lost2.End, common.Start, common.End, prevSelection.Start, prevSelection.End, selection.Start, selection.End );
		// Range.Difference() does not return the difference
		// when one range contains the other. So if prevSelection
		// contains the new selection the results must be fixed. 
		/*Range r=new Range(prevSelection);
		r.Sort();
		if (added.IsEmpty() && r.Contains(common.Start) && r.Contains(common.End)) {
			added.Start=common.Start;
			added.End=common.End;
			common.Clear();
		}*/	
		//System.Console.WriteLine("Common: {0} {1}",common.Start, common.End);
		//System.Console.WriteLine("Lost: {0} {1}",lost.Start, lost.End);
		//System.Console.WriteLine("Added: {0} {1}",added.Start, added.End);
			
		// render the lost parts normally
		int rLost1Start=-1;
		int rLost1End=-1;
		int rLost2Start=-1;
		int rLost2End=-1;
		
		if (!lost1.IsEmpty()) {
			rLost1Start=((int)(lost1.Start-offset)/bpr);
			rLost1End=((int)(lost1.End-offset)/bpr);
			//render the lines fully
			for (int j=rLost1Start; j<rLost1End; j++)
				RenderRow(j, 0, bpr, true, false);
			// make sure we are ot out of range at the last line
			long len = byteBuffer.Size-(lost1.End/bpr)*bpr;
            if (len > bpr) len=bpr;
            	RenderRow(rLost1End, 0, (int)len, true, false);

		}
	
		if (!lost2.IsEmpty()) {
			rLost2Start=((int)(lost2.Start-offset)/bpr);
			rLost2End=((int)(lost2.End-offset)/bpr);
			//render the middle lines fully
			for (int j=rLost2Start; j<rLost2End; j++)
				RenderRow(j, 0, bpr, true, false);
			// make sure we are ot out of range at the last line
			long len = byteBuffer.Size-(lost2.End/bpr)*bpr;
            if (len > bpr) len=bpr;
            	RenderRow(rLost2End, 0, (int)len, true, false);

		}
		
		
		// re-render the common parts that overlap the lost
		// parts as selections
		if (!common.IsEmpty()) {
			int rCommonStart=((int)(common.Start-offset)/bpr);
			int rCommonEnd=((int)(common.End-offset)/bpr);
			if (rCommonStart == rLost1End && rCommonEnd == rLost1End
				&& rLost1End == rLost1Start) {
				RenderSelection(common);
			}
			else if (rCommonStart == rLost2End && rCommonEnd == rLost2End
				&& rLost2End == rLost2Start) {
				RenderSelection(common);
			}
			else if (rCommonStart == rLost1End || rCommonStart == rLost2End) {
				int nbyte=(int)((common.Start-offset)%bpr);
				long overLeft = common.End-common.Start < bpr-nbyte ? common.End-common.Start:bpr-nbyte;
				Range rOverlap=new Range(common.Start, common.Start+overLeft);
				RenderSelection(rOverlap);
			} 
			else if (rCommonEnd == rLost1Start || rCommonEnd == rLost2Start) {
				int nbyte=(int)((common.End-offset)%bpr);
				long rCommonEndOffset=rCommonEnd*bpr+offset;
				long overStart=rCommonEndOffset > common.Start? rCommonEndOffset:common.Start;
				// part of common that is on row rCommonEnd
				Range rOverlap=new Range(overStart, rCommonEndOffset+nbyte);
				RenderSelection(rOverlap);
			}
		}
		
		
		// render the added parts as selections
		if (!added.IsEmpty()) {
			if ((added.Start==common.End+1) && added.Start>0)
				--added.Start;
			if (added.End==common.Start-1)
				++added.End;
			RenderSelection(added);
		}
		
		if (!added1.IsEmpty()) {
			if ((added1.Start==common.End+1) && added1.Start>0)
				--added1.Start;
			if (added1.End==common.Start-1)
				++added1.End;
			RenderSelection(added1);
		}
		
			
  	}
  	
  	///<summary>
	/// Render the bytes in range sel as selected
	///</summary>
 	protected void RenderSelection(Range sel)
 	{
 		int rstart, bstart, xstart, ystart;
 		int rend, bend, xend, yend;
 		bool odd;
 		Gdk.GC gc;
 		
 		
 		GetDisplayInfoByOffset(sel.Start,out rstart, out bstart, out xstart, out ystart);
 		GetDisplayInfoByOffset(sel.End,out rend, out bend, out xend, out yend);
		//System.Console.WriteLine("Start {0:x} {1} {2}", sel.Start, rstart, bstart);
 		//System.Console.WriteLine("End {0:x} {1} {2}", sel.End, rend, bend);
		if (rstart==rend) {
 			// odd row?
 			odd=(((sel.Start/bpr)%2)==1);
 			if (odd)
 				gc=backrOddcSelGC;
 			else	
 				gc=backrEvencSelGC;
 			//render
 			backPixmap.DrawRectangle(gc, true, xstart, ystart, xend-xstart, font.Height);
 			RenderRow(rstart, bstart, bend-bstart+1, false, true);
 		}
 		else {
 			//first
 			odd=(((sel.Start/bpr)%2)==1);
 			if (odd)
 				gc=backrOddcSelGC;
 			else	
 				gc=backrEvencSelGC;
 			backPixmap.DrawRectangle(gc, true, xstart, ystart, width-xstart, font.Height);
 			RenderRow(rstart, bstart, bpr-bstart, false, true);
 			
 			long curOffset=sel.Start+bpr-bstart;
 			//rest
 			for(int i=rstart+1;i<rend;i++) {
 				odd=(((curOffset/bpr)%2)==1);
 				if (odd) 
 					gc=backrOddcSelGC;
 				else
 					gc=backrEvencSelGC;
 				backPixmap.DrawRectangle(gc, true, 0, i*font.Height, width, font.Height);
 				RenderRow(i, 0, bpr, false, true);
 				curOffset+=bpr;
 			}
 			
 			//last
 			odd=(((sel.End/bpr)%2)==1);
 			if (odd)
 				gc=backrOddcSelGC;
 			else	
 				gc=backrEvencSelGC;
 			backPixmap.DrawRectangle(gc, true, 0, yend, xend, font.Height);
 			RenderRow(rend, 0, bend+1, false, true);
 		}
 		
 	}
	
	
	protected void RenderCursor()
	{
		// don't draw the cursor when there is a selection
		if (!selection.IsEmpty())
			return;
			
		int cRow, cByte, cX, cY;
 		GetDisplayInfoByOffset(cursorOffset, out cRow, out cByte, out cX, out cY);
 	
 		backPixmap.DrawRectangle(cursorGC, true, cX, cY+font.Height-2, font.Width*dpb, 2);
		if (cursorFocus) {
			backPixmap.DrawRectangle(cursorGC, true, cX+cursorDigit*font.Width, cY, 1, font.Height-2);
		} 	
	}
	
	public void DisposePixmaps()
	{
		backPixmap.Dispose();
		font.DisposePixmaps();
	}
	
}// Area


} //namespace
